/*
 * mibdump-xml.c --
 *
 *      Operations to dump SMIng module information as XML schemas.
 *
 * Copyright (c) 2002 T. Klie, Technical University of Braunschweig.
 *
 * See the file "COPYING" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * @(#) $Id: dump-xml.c,v 1.34 2004/07/06 10:41:18 tklie Exp $
 */

#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <netdb.h>

#include "smi.h"
#include "mibdump.h"
#include "table.h"

#ifdef HAVE_NET_SNMP_NET_SNMP_CONFIG_H
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#else
#include <ucd-snmp/ucd-snmp-config.h>
#include <ucd-snmp/ucd-snmp-includes.h>
#endif

#undef MD_USE_SNMP_VERSION_3

#ifdef MD_USE_SNMP_VERSION_3
#include "ucd-snmp/transform_oids.h"
const char *our_v3_passphrase = "The UCD Demo Password";
#endif

#define  INDENT		2    /* indent factor */
#define  INDENTVALUE	20   /* column to start values, except multiline */
#define  INDENTTEXTS	4    /* column to start multiline texts */
#define  INDENTMAX	64   /* max column to fill, break lines otherwise */


static char *schemaLocation = "http://www.ibr.cs.tu-bs.de/projects/libsmi/xsd/";
static int *dumpUnits = NULL;
static int *dumpDisplayHint = NULL;
static int *nestAugmentedTables = NULL;
static int *nestSubtables = NULL;

static struct snmp_session *ss;

static int silentFlag = 0;

typedef struct XmlEscape {
    char character;
    char *escape;
} XmlEscape;

static XmlEscape xmlEscapes [] = {
    { '<',	"&lt;" },
    { '>',	"&gt;" },
    { '&',	"&amp;" },
    { 0,	NULL }
};

static int current_column = 0;
static int ind = 1;

static int
min( int a, int b )
{
    return a < b ? a : b;
}

static int
numDigits( int val )
{
    int ret  = 1;

    for(; val / 10; val = val / 10 ) {
	ret++;
    }
    
    return ret;
}


static int
numOIDDigits( oid *anOID, int oid_len )
{
    int ret = 1;
    int pos;

    for( pos = 0; pos < oid_len; pos++ ) {
	if( ! anOID[ pos ] ) {
	    return ret;
	}
	ret += numDigits( anOID[ pos ] ) + 1;
    }

    return ret;
}

static char
*oidToStr( oid *anOID, int oid_len )
{
    char *ret = NULL;
    int pos;

    if( oid_len > 1 && !anOID[0] && !anOID[1] ) {
	/* zeroDotZero */
	return "0.0";
    }
    
    for( pos = 0; pos < oid_len; pos++ ) {
	if( !ret ) {
	    ret = xmalloc( numOIDDigits( anOID, oid_len ) + oid_len );
	    sprintf( ret, "%ld", anOID[ 0 ] );
	}
	else {
	    sprintf( ret, "%s.%ld", ret, anOID[ pos ] );
	}
    }

    return ret;
}



static char 
*getOctStr( struct variable_list *vars )
{
     int i;
     char *ret;

     ret = xmalloc( 3 * vars->val_len );
     for( i = 0; i < vars->val_len; i++ ) {
	 if( ! i ) {
	     sprintf( ret, "%02x", vars->val.string[0] );
	 }
	 else {
	     sprintf( ret, "%s%02x", ret, vars->val.string[i] );
	 }
     }
     return ret;
}


static char
*xmlEncode( const char *s )
{
    char *ret = NULL;
    unsigned int len;
    unsigned int i,j;
    
    if (s) {
	len = strlen(s);
	for (i=0; i < len; i++) {
	    for (j = 0; xmlEscapes[j].character; j++) {
		if (xmlEscapes[j].character == s[i]) break;
	    }
	    if (xmlEscapes[j].character) {
		if( ! ret ) {
		    ret = xmalloc( len + strlen( xmlEscapes[j].escape ) + 1 );
		    sprintf( ret, "%s", xmlEscapes[j].escape );
		}
		else {
		    ret = xrealloc( ret, strlen( ret ) + 1 +
				    len + strlen( xmlEscapes[j].escape ) );
		    sprintf( ret, "%s%s", ret, xmlEscapes[j].escape );
		}
	    } else {
		if( ! ret ) {
		    ret = xmalloc( len + 1 );
		    sprintf( ret, "%c", s[ i ] );
		}
		else {
		    sprintf( ret, "%s%c", ret, s[ i ] );
		}

	    }
	    ret[ strlen( ret ) ] = '\0';
	}
    }
    return ret;
}


#if 0
static char
*getAsciiStr( struct variable_list *vars )
{
    char *ret, *s;

    ret = xmalloc( vars->val_len + 1 );
    strncpy( ret, vars->val.string, vars->val_len );
    ret[ vars->val_len ] = '\0';

    s = xmlEncode( ret );
    xfree( ret );
    return s;
}

static int
isAsciiString( char *string, unsigned int stringlen )
{
  int i;

    for( i = 0; i < stringlen; i++ ) {
	if( ! isprint( string[i] ) ) {
	    /* value is non-printable */
	    return 0;
	}
    }

    return 1;
}

static int
containsAsciiString( struct variable_list *vars )
{
    return isAsciiString( vars->val.string, vars->val_len );
}


static unsigned int
getNamedNumberCount( SmiType *smiType )
{
    SmiNamedNumber *nn;
    unsigned int ret = 0;

    for( nn = smiGetFirstNamedNumber( smiType );
	 nn;
	 nn = smiGetNextNamedNumber( nn ) ) {
	ret++;
    }

    return ret;
}
#endif /* 0 */

static int
power( int a, unsigned int b )
{
    unsigned int i;
    int ret = 1;

    for( i = 0; i < b; i++ ) {
	ret *= a;
    }
    return ret;
}


static char
*renderInt( struct variable_list *vars, char *hint )
{
    char *ret = NULL;
    unsigned int j;
    
    switch( hint[ 0 ] ) {

    case 'd':
	/* render as decimal */
	if( strlen( hint ) > 2  ) {
	    unsigned int offset = atoi( &hint[2] );
	    asprintf( &ret,"%d.%d",
		      (int)vars->val.integer / power( 10, offset ), 
		      abs( (int)vars->val.integer % power( 10, offset ) ) );
	}
	else {
	    asprintf( &ret, "%d", vars->val.integer );
	}
	return ret;

    case 'b':
	/* render as binary */
	for( j = 7; j >= 0; j-- ) {
	    asprintf( &ret, "%s%d", ret ? ret : "",
		      ( (int)vars->val.integer &
			power( 2, j ) ) != 0 );
	}
	return ret;

    case 'o':
	/* render as octal value */
	asprintf( &ret, "%03o", vars->val.integer );
	return ret;

    case 'x':
	/* render as hex */
	asprintf( &ret, "%02x", vars->val.integer );
	return ret;	
    }
    return 0;
}

static char
*renderStr( unsigned char *strVal, size_t strVal_len, char *hint )
{
    int hintPos = 0, lastHintPos = 0;
    int varPos = 0;
    char *ret = NULL;
    int repeat = 0;
    int intNum = 0;  

    while( hintPos < strlen( hint ) && varPos < strVal_len ){

	switch( hint[ hintPos ] ) {

	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9': {  
	    unsigned int endPos;
		char *strNum;
		
		/* find end of number */
		for( endPos = hintPos; isdigit( hint[ endPos ] ); endPos++ );
		
		/* parse number */
		strNum = xmalloc( endPos - hintPos );
		strncpy( strNum, &hint[ hintPos ], endPos - hintPos );
		strNum[ endPos - hintPos ] = '\0';
		intNum= atoi( strNum );
		xfree( strNum );
		
		/* forward the position pointer */
		hintPos = endPos;
		break;
	}

	case 'a':
	case 't':
	    /* render as ascii */
	    if( repeat ) {
		/* xxx TBD */    
	    }
	    else if( intNum ) {
		/* lengths depends on number read last iteration */
		int len = min( strVal_len - varPos, intNum );
		char * strval = strndup( &strVal[ varPos ], len );

		if( ! strVal[ varPos ] ) {
		    return NULL;
		}
		
		asprintf( &ret, "%s%s", ret ? ret : "",
			  xmlEncode( strval ) );

		/* forward position pointers */
		varPos += len;
		hintPos++;		
	    }
	    else {
		/* xxx TBD */
	    }
	    break;

	case 'b':
	    /* render as binary */
	    /* xxx TBD: repeat, intNum = 0 */

	    if( intNum ) {
		int i,j;

		for( i = 0;
		     i < min( intNum, strVal_len - varPos + 1);
		     i++ ) {

		    for( j = 7; j >= 0; j-- ) {
			asprintf( &ret, "%s%d", ret ? ret : "",
				  ( strVal[ varPos ] &
				    power( 2, j ) ) != 0 );
		    }

		    /* forward the position pointer */
		    varPos++;
		}
	    }
	    /* forward the position pointer */
	    hintPos++;
	    break;

	case 'd':
	    /* render as decimal */
	    /* xxx TBD: repeat, intNum = 0 */
	    if( intNum ) {
		int strNum_len = min( intNum, strVal_len - varPos + 1 );
		int intVal = 0, i;
		
		/* xxx check byte order */
		for( i = 0; i < strNum_len; i++ ) {
		    intVal = intVal * 256 +
			(unsigned int)strVal[ i + varPos ];
		}	       
		asprintf( &ret, "%s%d", ret ? ret : "", intVal );
		varPos += strNum_len;
	    }
	    /* forward position pointer */
	    hintPos++;
	    break;

	case 'o':
	    /* render as octal */
	    /* xxx TBD: repeat, intNum = 0 */
	    if( intNum ) {
		int octVal_len = min( intNum, strVal_len - varPos );
		int i;

		for( i = 0; i < octVal_len; i++) {
		    asprintf( &ret, "%s%03o",
			      ret ? ret : "", 
			      strVal[ i ] );
		}
		/* forward the position pointer */
		varPos += octVal_len;
		hintPos++;
	    }
	    break;

	case 'x':
	    /* render as hex */
	    /* xxx TBD: repeat, intNum = 0 */
	    if( intNum ) {
		int hexVal_len = min( intNum, strVal_len - varPos );
		int i;

		for( i = varPos; i < hexVal_len + varPos; i++) {
		    asprintf( &ret, "%s%02x",
			      ret ? ret : "", 
			      strVal[ i ] );
		}
		/* forward the position pointer */
		varPos += hexVal_len;
		hintPos++;
		
	    }
	    break;
	default:
	    /* use the character in the hint as a delimeter */
	    if( varPos < strVal_len ) {
		asprintf( &ret, "%s%c",
			  ret ? ret : "", hint[ hintPos++ ] );
	    }
	    break;
	}
	if( hintPos >= strlen( hint ) ) {
	    hintPos = lastHintPos;
	}
	else if( isdigit( hint[ hintPos ] ) ){
	    lastHintPos = hintPos;
	}
    }
    return ret;
}


static char*
getStringVal( struct variable_list *vars, SmiNode *smiNode )
{
    char* ret = NULL;
    SmiType *smiType = smiNode ? smiGetNodeType( smiNode ) : NULL;  

    if( smiType ) {
	int error = 0;
	
	switch( smiType->basetype ) {
	    
	case SMI_BASETYPE_INTEGER32: {
	    SmiType *parentType = smiGetParentType( smiType );
	    
	    if( vars->type != ASN_INTEGER ) {
		fprintf( stderr,
			 "mibdump: warning: type conflict bewteen mib "
			 "and agent. Node `%s' is supposed to have basetype "
			 "integer32.\n", smiNode->name );
	    }
	    if( smiType->format ) {
		return renderInt( vars, smiType->format );
	    }
	    else if( parentType && parentType->format ) {
		return renderInt( vars, parentType->format );
	    }
	    else {
		ret = xmalloc( numDigits( *(vars->val.integer) ) );
		sprintf( ret, "%ld", *(vars->val.integer) );
		return ret;
	    }
	}

	    
	case SMI_BASETYPE_OCTETSTRING:
	    if( vars->type != ASN_OCTET_STR ) {
		if( ( ! ( ! strcmp( smiType->name, "IpAddress" ) ) &&
		      ( vars->type == ASN_IPADDRESS ) ) ){
		    fprintf( stderr,
			     "mibdump: warning: type conflict bewteen mib "
			     "and agent. "
			     "Node `%s' is supposed to have basetype "
			     "octet string\n", smiNode->name );
		}
	    }
	    if( vars->val_len ) {
		SmiType *smiType = smiGetNodeType( smiNode );
		SmiType *parentType = smiGetParentType( smiType );
		if( smiType->format ) {
		    return renderStr( vars->val.string, vars->val_len,
				      smiType->format );
		}
		else if( parentType && parentType->format ) {
		    return renderStr( vars->val.string, vars->val_len,
				      parentType->format );
		}
		else {
		    return getOctStr( vars );
		}
		
	    }
	    else {
		ret = NULL;
	    }
	    return ret;
	    
	case SMI_BASETYPE_OBJECTIDENTIFIER:
	      if( vars->type != ASN_OBJECT_ID ) {
		  fprintf( stderr,
			   "mibdump: warning: type conflict bewteen mib "
			   "and agent. Node `%s' is supposed to have basetype "
			   "object identifier\n", smiNode->name );
	      }
	      return oidToStr( vars->val.objid,
			       vars->val_len / sizeof( oid ) );
	    
	case SMI_BASETYPE_UNSIGNED32:
	    switch( vars->type ) {
		
	    case ASN_COUNTER:
		if( smiType->name && strcmp( smiType->name, "Counter32" ) ){
		    error = 0;
		}
		break;

	    case ASN_GAUGE:
		if(  smiType->name && strcmp( smiType->name, "Gauge32" ) ) {
		    error = 0;
		}
		break;

	    case ASN_TIMETICKS:
		if(  smiType->name && strcmp( smiType->name, "TimeStamp" ) ) {
		    error = 0;
		}
		break;

	    case ASN_INTEGER:
		error = 0;
		break;

	    default:
		error = 1;
		break;
	    }
	    if( error ) {
		fprintf( stderr,
			 "mibdump: warning: type conflict bewteen mib "
			 "and agent. "
			 "Node `%s' `%x' is supposed to have basetype "
			 "unsigned32.\n", smiNode->name, vars->type );
	    }
	    
	    ret = xmalloc( numDigits( *(vars->val.integer) ) );
	    sprintf( ret, "%lu", *(vars->val.integer) );
	    return ret;
	    
	case SMI_BASETYPE_INTEGER64:
	    if( vars->type != ASN_INTEGER ) {
		fprintf( stderr,
			 "mibdump: warning: type conflict bewteen mib "
			 "and agent. Node `%s' is supposed to have basetype "
			 "integer64.\n", smiNode->name );
	    }
	    ret = xmalloc( numDigits( *(vars->val.integer) ) );
	    sprintf( ret, "%ld", *(vars->val.integer) );
	    return ret;
	    
	case SMI_BASETYPE_UNSIGNED64:
	    if( vars->type != ASN_INTEGER ) {
		fprintf( stderr,
			 "mibdump: warning: type conflict bewteen mib "
			 "and agent. Node `%s' is supposed to have basetype "
			 "unsigned64.\n", smiNode->name );
	    }
	    ret = xmalloc( numDigits( *(vars->val.integer) ) );
	    sprintf( ret, "%lu", *(vars->val.integer) );
	return ret;
	
	case SMI_BASETYPE_ENUM: {
	    SmiNamedNumber *namedNumber;
	    
	    if( vars->type != ASN_INTEGER ) {
		fprintf( stderr,
			 "mibdump: warning: type conflict bewteen mib "
			 "and agent. Node `%s' is supposed to have basetype "
			 "integer.\n", smiNode->name );
	    }
	    for( namedNumber = smiGetFirstNamedNumber( smiType );
		 namedNumber;
		 namedNumber = smiGetNextNamedNumber( namedNumber ) ){
		if( namedNumber->value.value.integer32 == *(vars->val.integer) ) {
		    return namedNumber->name;
		}
	    }
	    fprintf( stderr,
		     "mibdump: warning: number `%ld' not in enumerarion "
		     "for node `%s'\n",
		     *(vars->val.integer), smiNode->name );
	    return NULL;
	}
	    
	case SMI_BASETYPE_BITS:
	{
	    SmiNamedNumber *namedNumber;
	    
	    if( vars->type != ASN_OCTET_STR ) {
		fprintf( stderr,
			 "mibdump: warning: type conflict bewteen mib "
			 "and agent. Node `%s' is `%d'supposed to have basetype "
			 "bits.\n", smiNode->name , vars->type);
	    }
	    for( namedNumber = smiGetFirstNamedNumber( smiType );
		 namedNumber;
		 namedNumber = smiGetNextNamedNumber( namedNumber ) ){

		if( vars->val.
		    bitstring[ namedNumber->value.value.unsigned32 / 8 ] &
		    ( 128 /  power( 2,
				    namedNumber->value.value.unsigned32 % 8
			) ) ) {

		    if( ret ) {
			ret = xrealloc( ret, strlen( ret ) +
					strlen( namedNumber->name ) + 2 );
			sprintf( ret, "%s %s", ret, namedNumber->name );
		    }
		    else {
			ret = xmalloc( strlen( namedNumber->name + 1 ) );
			strcpy( ret, namedNumber->name );
		    }
		}
	    }
	    return ret;
	}
	default : return "basetype not implemented yet";
	}
    }

    if( ! smiNode ) {
	fprintf( stderr,
		 "mibdump: warning: found unknown object in MIB: `%s'\n",
		 oidToStr( vars->name, vars->name_length ) );
	return NULL;
    }
    return ret;
}

static int numIndex( SmiNode *smiNode )
{
    SmiElement *iterElem;
    int ret = 0;

    for( iterElem = smiGetFirstElement( smiNode );
	 iterElem;
	 iterElem = smiGetNextElement( iterElem ) ) {
	ret++;
    }
    return ret;
}


static struct snmp_session *initSnmp()
{
    struct snmp_session session, *ss;
    /* initialize SNMP library */
    init_snmp( "mibdump" );

    /*
     * Initialize a "session" that defines who we're going to talk to
     */
    snmp_sess_init( &session );                   /* set up defaults */
    session.peername = snmpHost;
    session.remote_port = snmpPort;

    /* set up the authentication parameters for talking to the server */
    
#ifdef MD_USE_SNMP_VERSION_3
    
    /* Use SNMPv3 to talk to the experimental server */
    
    /* set the SNMP version number */
    session.version=SNMP_VERSION_3;
    
    /* set the SNMPv3 user name */
    session.securityName = strdup("MD5User");
    session.securityNameLen = strlen(session.securityName);
    
    /* set the security level to authenticated, but not encrypted */
    session.securityLevel = SNMP_SEC_LEVEL_AUTHNOPRIV;
    
    /* set the authentication method to MD5 */
    session.securityAuthProto = usmHMACMD5AuthProtocol;
    session.securityAuthProtoLen = sizeof(usmHMACMD5AuthProtocol)/sizeof(oid);
    session.securityAuthKeyLen = USM_AUTH_KU_LEN;
    
    /* set the authentication key to a MD5 hashed version of our
       passphrase "The UCD Demo Password" (which must be at least 8
       characters long) */
    if (generate_Ku(session.securityAuthProto,
                    session.securityAuthProtoLen,
                    (u_char *) our_v3_passphrase, strlen(our_v3_passphrase),
                    session.securityAuthKey,
                    &session.securityAuthKeyLen) != SNMPERR_SUCCESS) {
        snmp_perror(argv[0]);
        snmp_log(LOG_ERR,
                 "Error generating Ku from authentication pass phrase. \n");
        exit(1);
    }
    
#else /* we'll use the insecure (but simplier) SNMPv1 */
    
    /* set the SNMP version number */
    session.version = SNMP_VERSION_1;
    
    /* set the SNMPv1 community name used for authentication */
    session.community = snmpCommunity;
    session.community_len = strlen(session.community);
    
#endif /* SNMPv1 */

    /* windows32 specific initialization (is a noop on unix) */
    SOCK_STARTUP;
    
    /*
     * Open the session
     */
    ss = snmp_open(&session);                     /* establish the session */

    if (!ss) {
        snmp_perror("ack");
        snmp_log(LOG_ERR, "something horrible happened!!!\n");
	
	fprintf( stderr, "host: %s, community: %s\n", snmpHost, snmpCommunity );
        exit(2);
    }
    
    return ss;
}


static void
fprint(FILE *f, char *fmt, ...)
{
    va_list ap;
    char    s[200];
    char    *p;

    va_start(ap, fmt);
#ifdef HAVE_VSNPRINTF
    current_column += vsnprintf(s, sizeof(s), fmt, ap);
#else
    current_column += vsprintf(s, fmt, ap);	/* buffer overwrite */
#endif
    va_end(ap);

    fputs(s, f);

    if ((p = strrchr(s, '\n'))) {
	current_column = strlen(p) - 1;
    }
}




static void fprintSegment(FILE *f, int relindent, char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
     
    if ((ind == 0) || (ind + relindent == 0)) {
	ind += relindent;
    } else {
	if (relindent < 0) ind += relindent;
	fprintf(f, "%*c", ind * INDENT, ' ');
	if (relindent > 0) ind += relindent;
    }
    vfprintf(f, fmt, ap);

    va_end(ap);
}



#if 0
static void
fprintMultilineString(FILE *f, int column, const char *s)
{
    int i, j, len;

#ifdef INDENTTEXTS
    fprintSegment(f, column + INDENTTEXTS, "", 0);
#endif /* #ifdef INDENTTEXTS */
    if (s) {
	len = strlen(s);
	for (i=0; i < len; i++) {
	    for (j = 0; xmlEscapes[j].character; j++) {
		if (xmlEscapes[j].character == s[i]) break;
	    }
	    if (xmlEscapes[j].character) {
		fputs(xmlEscapes[j].escape, f);
		current_column += strlen(xmlEscapes[j].escape);
	    } else {
		putc(s[i], f);
		current_column++;
	    }
	    if (s[i] == '\n') {
		current_column = 0;
#ifdef INDENTTEXTS
		fprintSegment(f, column + INDENTTEXTS, "", 0);
#endif /* #ifdef INDENTTEXTS */
	    }
	}
    }
    current_column++;
}
#endif /* #if 0 */


static int
hasChildren( SmiNode *smiNode, SmiNodekind nodekind )
{
    SmiNode *iterNode;
    int childNodeCount = 0; 

    for( iterNode = smiGetFirstChildNode( smiNode );
	 iterNode;
	 iterNode = smiGetNextChildNode( iterNode ) ){
	if( nodekind & iterNode->nodekind ) {
	    childNodeCount++;
	}
    }
    return childNodeCount;
}


static oid *
oidcpy( SmiNode *smiNode )
{
    oid *anOID;
    int j;
    
    anOID = xmalloc( smiNode->oidlen * sizeof( oid ) );
    for( j = 0; j < smiNode->oidlen; j++ ) {
	anOID[j] = (oid)smiNode->oid[j];
    }
    return anOID;
}


static oid *
oidadd( oid *anOID, int oidlen, oid *suffix, int suflen, int *newlen )
{
    int j;
    
    anOID = xrealloc( anOID, ( oidlen + suflen ) * sizeof( oid ) );
    for( j = oidlen; j - oidlen < suflen; j++ ) {
	anOID[ j ] = suffix[ j - oidlen ];
    }
    *newlen = oidlen + suflen;
    return anOID;
}


static int
snmpGet( oid *anOID,
	 int oidlen,
	 u_char method,
	 struct snmp_pdu **response )
{
    struct snmp_pdu *pdu = NULL;
    int status;

    if( ! ss ) {
	ss = initSnmp();
    }
    
    pdu = snmp_pdu_create( method );
    
    snmp_add_null_var( pdu, anOID, oidlen );
    status = snmp_synch_response( ss, pdu, response );

    if( status == STAT_SUCCESS && (*response)->errstat == SNMP_ERR_NOERROR) {
	return 1;
    }
    else {
	int i;
	
	fprintf( stderr, "NOT READY:\n" );
	for( i = 0; i < oidlen; i++ ) {
	    fprintf( stderr, ".%lu",anOID[i] );
	}
	fprintf( stderr, "\n" );
	if (status == STAT_SUCCESS)
	    fprintf(stderr, "Error in packet\nReason: %s\n",
		    snmp_errstring( (*response)->errstat));
	else
	    snmp_sess_perror("snmpget", ss);
	return 0;
    }
}

static SmiType
*findTypeWithRange(SmiType *smiType)
{
    SmiType *iterType;

    for (iterType = smiType; iterType; iterType = smiGetParentType(iterType)) {
	if (smiGetFirstRange(iterType)) {
	    return iterType;
	}
    }
    return NULL;
}


static int
isFixedLengthString( SmiType *smiType, int *fixedLength )
{
    SmiType *twr = findTypeWithRange( smiType );
    SmiInteger32 minLength = -1, maxLength = -1;
    SmiRange *iterRange;

    if( ! twr ) {
	return 0;
    }

    for( iterRange = smiGetFirstRange( twr );
	 iterRange;
	 iterRange = smiGetNextRange( iterRange ) ) {
	if( minLength == -1 ) {
	    minLength = iterRange->minValue.value.integer32;
	}
	else if( iterRange->minValue.value.integer32 != minLength ) {
	    return 0;
	}
	if( maxLength == -1 ) {
	    maxLength = iterRange->maxValue.value.integer32;
       	}
	else if ( iterRange->maxValue.value.integer32 != maxLength ) {
	    return 0;
	}
	if( maxLength != minLength ) {
	    return 0;
	}
    }
    *fixedLength = (int)minLength;
    return 1;
}


static char *
decodeStrValFromOID( SmiNode *smiNode, oid *anOID, unsigned int oidlen,
		     unsigned int offset )
{
    SmiElement *iterElem;
    char *ret = NULL;
    unsigned int oidpos = 0;
    unsigned int iterpos = 0;
    
    for( iterElem = smiGetFirstElement( smiNode );
	 iterElem;
	 iterElem = smiGetNextElement( iterElem ) ) {
	SmiNode *iterNode = smiGetElementNode( iterElem );
	SmiType *iterType = smiGetNodeType( iterNode );

	if( iterpos++ < offset ) {
	    continue;
	}

	switch( iterType->basetype ) {
	    
	case SMI_BASETYPE_UNSIGNED64:
	case SMI_BASETYPE_INTEGER64:
	case SMI_BASETYPE_UNSIGNED32:
	case SMI_BASETYPE_INTEGER32:
	    if( ret ) {
		asprintf( &ret, "%s ", ret );
	    }
	    asprintf( &ret, "%s%s=\"%lu\"",
		      ret ? ret : "",
		      iterNode->name, anOID[ oidpos ] );
	    
	    oidpos++;
	    break;
	    
	case SMI_BASETYPE_OBJECTIDENTIFIER: {
	    /* xxx TBD */
	    break;
	}
	    
	case SMI_BASETYPE_ENUM: {
	    SmiNamedNumber *nn;
	    
	    for( nn = smiGetFirstNamedNumber( iterType );
		 nn;
		 nn = smiGetNextNamedNumber( nn ) ) {
		if( nn->value.value.integer32 == anOID[ oidpos ] ) {
		    /* named number found ==> exit loop */
		    break;
		}
	    }

	    if( !nn ) {
		/* no named number found */
		fprintf( stderr, "mibdump: warning: no named number found\n" );
		oidpos++;
		break;
	    }
	    
	    if( ret ) {
		asprintf( &ret, "%s ", ret );
	    }
	    asprintf( &ret, "%s%s=\"%s\"",
		      ret ? ret : "", iterNode->name, nn->name );
	    
	    oidpos++;
	    break;
	    
	}
	case SMI_BASETYPE_BITS:
	    /* xxx TBD */
	    break;
	    
	case SMI_BASETYPE_OCTETSTRING:
	{
	    unsigned int i, fixedLength, len;
	    char *subIdxStr = NULL;
	    SmiType *parentType = smiGetParentType( iterType );

	    len = isFixedLengthString( iterType, &fixedLength ) ?
		fixedLength : (unsigned int)anOID[ oidpos++ ];
	    
	    if( iterType->format ) {
		/* render using display hint */
		subIdxStr = renderStr( (unsigned char*)&anOID[ oidpos ], len,
				       iterType->format );
		}
	    else if( parentType && parentType->format ) {
		/* render using parent display hint */
		subIdxStr = renderStr( (unsigned char*)&anOID[ oidpos ], len,
				       parentType->format );  
	    }
	    else if( !strcmp( iterType->name, "IpAddress" ) ) {
		/* SNMPv2-SMI:IpAddress does not have a display hint.
		   ==> let's use this one: "1d." */
		subIdxStr = renderStr( (unsigned char*)&anOID[ oidpos ], len,
				       "1d." );
	    }
	    else {
		/* render normally */
		for( i = oidpos; i < oidpos + len; i++ ) {
		    asprintf( &subIdxStr, "%s%c",
			      subIdxStr ? subIdxStr : "", anOID[ i ] );
		}
	    }
	    if( ret ) {
		asprintf( &ret, "%s %s=\"%s\"", ret,
			  iterNode->name, subIdxStr );
	    }
	    else {
		asprintf( &ret, "%s=\"%s\"", iterNode->name, subIdxStr );
	    }			           
	    
	    oidpos += len;
	    break;
	}
		
	case SMI_BASETYPE_FLOAT32:
	case SMI_BASETYPE_FLOAT64:
	case SMI_BASETYPE_FLOAT128:
	/* xxx TBD: (only SMIng) */
	    break;
	    
	case SMI_BASETYPE_UNKNOWN:
	    /* should not occur */
	    if( smiNode ) {
		if( smiNode->name ) {
		    fprintf( stderr,
			     "mibdump: warning: unknown basetype "
			     "in node `%s'\n", smiNode->name );
		}
		else {
		    fputs( "mibdump: warning: unknown basetype\n", stderr );
		}
	    }
	    else {
		fputs( "mibdump: warning: unknown basetype\n", stderr );
	    }
	}
    }
    return ret;
} 


static char *
getIndexStr( SmiNode *tableNode, MDRow *row,
	     oid *rowIndex, size_t rowIndex_len, unsigned int offset )
{
    char *ret = NULL;

    switch( tableNode->indexkind ) {

    case SMI_INDEX_INDEX:
	if( rowIndex && !memcmp( rowIndex, row->index, rowIndex_len ) ) {
	    /* ignore parent index when dumping nested subtables */
	    ret = decodeStrValFromOID( tableNode, &row->index[ rowIndex_len ],
				       row->index_len - rowIndex_len, offset );
	}
	else {
	    ret = decodeStrValFromOID( tableNode, row->index, row->index_len,
				       0 );
	}
	break;
	
    case SMI_INDEX_AUGMENT: 
	ret = decodeStrValFromOID( smiGetRelatedNode( tableNode ),
				   row->index, row->index_len, 0 );
	break;

    case SMI_INDEX_REORDER:
    case SMI_INDEX_SPARSE:
    case SMI_INDEX_EXPAND:
	/* only SMIng */
	break;

    case SMI_INDEX_UNKNOWN:
	if( tableNode ) {
	    if( tableNode->name ) {
		fprintf( stderr, "mibdump: error: unknown index method "
			 "in table `%s'\n", tableNode->name );
	    }
	    else {
		fputs( "mibdump: error: unknown index method\n", stderr );
	    }
	}
	else {
	    fputs( "mibdump: error: unknown index method\n", stderr );
	}
    }

    return ret;
}


static SmiNode *
getColumnNode( SmiNode *tableNode, oid col )
{
    SmiNode *iterNode;

    for( iterNode = smiGetFirstChildNode( tableNode );
	 iterNode;
	 iterNode = smiGetNextChildNode( iterNode ) ) {
	if( iterNode->oid[ iterNode->oidlen - 1 ] == (SmiSubid)col ) {
	    return iterNode;
	}
    }
    return NULL;
}

static void
fprintColumns( FILE *f, MDRow *row, SmiNode *tableNode,
	       MDTable *table )
{
    MDCol *iterCol;
    MDAugment *actAugment = table->augmentations;
    
    for( iterCol = row->columns; iterCol; iterCol = iterCol->next ) {
	SmiElement *iterElem;
	int cont = 1;
	SmiNode *colNode = getColumnNode( tableNode, iterCol->col  );
	
	if( ! colNode ) {
	    actAugment = md_getAugment( table, iterCol );
	    if( actAugment ) {
		colNode =
		    getColumnNode( actAugment->smiNode,
				   iterCol->col - actAugment->startCol );
	    }
	}
	
	/* check if we are index column */
	for( iterElem = smiGetFirstElement( smiGetParentNode( colNode ) ) ;
	     iterElem;
	     iterElem = smiGetNextElement( iterElem ) ) {
	    if( colNode == smiGetElementNode( iterElem ) ) {
		/* we are index coulumn ==> do not print element */
		cont = 0;
		break;
	    }
	}
	
	/* normal column */
	if( colNode && cont && iterCol->value && ! iterCol->subRow ) {
	    SmiType *colType = smiGetNodeType( colNode );
	    
	    fprintSegment( f, 0, "<%s", colNode->name );
	    if( dumpUnits && colNode->units ) {
		fprint( f, " units=\"%s\"", colNode->units );
	    }
	    if( dumpDisplayHint ) {
		SmiType *colParentType = smiGetParentType( colType );
		char *format = NULL;
		
		if( colType->format ) {
		    format = colType->format;
		}
		else if( colParentType && colParentType->format ) {
		    format = colParentType->format;
		}
		if( format ) {
		    fprint( f, " displayHint=\"%s\"", format );
		}		    
	    }
	    fprint( f, ">%s</%s>\n",
		    iterCol->value, colNode->name );		
	}
	
	/* subtable columns */	    
	else if( iterCol->subRow ) {
	    MDRow *iterSubRow;
	    
	    for( iterSubRow = iterCol->subRow;
		 iterSubRow &&
		     ( iterSubRow->index_len == iterCol->subRow->index_len ) &&
		     !memcmp( iterSubRow->index, iterCol->subRow->index,
			      iterSubRow->index_len );
		 iterSubRow = iterSubRow->next ) {
		fprintSegment( f, 1, "<%s %s>\n",
			       iterCol->value,
			       getIndexStr(
				   iterCol->smiNode , iterSubRow,
				   row->index, row->index_len,
				   numIndex( iterCol->smiNode ) - 1 ) );
		fprintColumns( f, iterSubRow, iterCol->smiNode, table );
		fprintSegment( f, -1, "</%s>\n",  iterCol->value );
	    }
	}
    }
}


static void
fprintTable( FILE *f, MDTable *table, SmiNode *tableNode )
{
    MDRow *iterRow;
    
    for( iterRow = table->rows; iterRow; iterRow = iterRow->next ) {
	char *indexStr = getIndexStr( tableNode, iterRow, NULL, 0, 0 );


	if( md_numColumns( iterRow ) > numIndex( tableNode ) ) {
	    /* print row with index */
	    fprintSegment( f, 1, "<%s %s>\n", table->name, indexStr );
	    xfree( indexStr );
	    
	    /* print all columns */
	    fprintColumns( f, iterRow, tableNode, table );
	    
	    fprintSegment( f, -1, "</%s>\n", table->name );
	}
	else {
	     fprintSegment( f, 0, "<%s %s/>\n", table->name, indexStr );
	}
    }
}

static void
fillTable( MDTable *table, SmiNode *smiNode, unsigned int colOffset ) {
    MDRow *row;
    oid col;
    oid *prefix = oidcpy( smiNode );
    size_t prefix_len = smiNode->oidlen;
    SmiNode *actNode = NULL;
    struct snmp_pdu *response;
    MDAugment *iterAugm;

    if( colOffset ) {
	/* we are an augmentation */
	if( !table->augmentations ) {
	    table->augmentations = xmalloc( sizeof( struct MDAugment ) );
	    table->augmentations->startCol = colOffset;
	    table->augmentations->smiNode = smiNode;
	    table->augmentations->next = NULL;
	}
	else {
	    for( iterAugm = table->augmentations;
		 iterAugm->next;
		 iterAugm = iterAugm->next );
	    
	    iterAugm->next = xmalloc( sizeof( struct MDAugment ) );
	    iterAugm->next->startCol = colOffset;
	    iterAugm->next->smiNode = smiNode;
	    iterAugm->next->next = NULL;
	}
    }
    
    if( snmpGet( prefix, prefix_len, 
		 SNMP_MSG_GETNEXT, &response ) ) {
	while( ! memcmp( response->variables->name, prefix,
			 MIN( response->variables->name_length,
			      prefix_len ) * sizeof( oid ) ) ) {
	    oid newOID[MAX_OID_LEN];
	    int newOID_len;
	    oid *idx;
	    char *strval;
	    MDCol *mdcol;
	    
	    col = response->variables->name[ prefix_len ];
	    idx = &(response->variables->name[prefix_len + 1]);
	    row = md_findRow( table, idx,
			      response->variables->name_length -
			      prefix_len - 1 );
	    actNode = getColumnNode( smiNode, col );
	    
	    if( !row ) {
		/* row does not exist */
		row = md_addRow( table, idx,
				 response->variables->name_length -
				 prefix_len - 1 );
	    }
	    strval = getStringVal( response->variables, actNode );
	    mdcol = md_addColumn( row, col + colOffset,strval );
	    table->maxColNum = MAX( table->maxColNum, col + colOffset );
	    
	    /* get next value */
	    memcpy( newOID, response->variables->name,
		    response->variables->name_length * sizeof( oid ) );
	    newOID_len = response->variables->name_length;
	    if( ! snmpGet( newOID, newOID_len,
			   SNMP_MSG_GETNEXT, &response ) ) {
		break; /* if we cannot get more values let's exit loop */
	    }
	}
    }
}

static oid
*extractIndex( SmiNode *smiNode, SmiNode *idxNode, size_t *extract_len,
	       oid *anOID, size_t anOID_len )
{
    oid *ret;
    SmiElement *iterElem;
    unsigned int oidpos = 0;
    
    /* extract index oid part */
    for( iterElem = smiGetFirstElement( smiNode );
	 iterElem;
	 iterElem = smiGetNextElement( iterElem ) ) {
	SmiNode *iterNode = smiGetElementNode( iterElem );
	SmiType *iterType = smiGetNodeType( iterNode );
	int startPos = 0, endPos;
	
	if( iterNode == idxNode ) {
	    startPos = oidpos;
	}
	switch( iterType->basetype ) {
	    
	case SMI_BASETYPE_UNSIGNED64:
	case SMI_BASETYPE_INTEGER64:
	case SMI_BASETYPE_UNSIGNED32:
	case SMI_BASETYPE_INTEGER32:
	case SMI_BASETYPE_ENUM:
	    oidpos++;
	    break;
	    
	case SMI_BASETYPE_OBJECTIDENTIFIER:
	{
	    /* xxx TBD */
	    break;
	}
	
	case SMI_BASETYPE_BITS:
	{
	    /* xxx TBD */
	    break;
	}
	
	case SMI_BASETYPE_OCTETSTRING:
	{
	    unsigned int fixedLength, len;
	    
	    len = isFixedLengthString( iterType, &fixedLength ) ?
		fixedLength : (unsigned int)anOID[ oidpos++ ];	    
	    oidpos += len;
	    break;
	}
	
	case SMI_BASETYPE_FLOAT32:
	case SMI_BASETYPE_FLOAT64:
	case SMI_BASETYPE_FLOAT128:
	    /* xxx TBD: (only SMIng) */
	    break;
	    
	case SMI_BASETYPE_UNKNOWN:
	    /* should not occur */
	    if( smiNode ) {
		if( smiNode->name ) {
			fprintf( stderr,
				 "mibdump: warning: unknown basetype "
				 "in node `%s'\n", smiNode->name );
		}
		else {
		    fputs( "mibdump: warning: unknown basetype\n", stderr );
		}
	    }
	    else {
		fputs( "mibdump: warning: unknown basetype\n", stderr );
	    }
	    break;
	    
	}
	
	if( iterNode == idxNode ) {
	    unsigned int i;
	    
	    endPos = oidpos;
	    *extract_len = endPos - startPos;
	    ret = xmalloc( *extract_len * sizeof( oid ) );
/*	    memcpy( ret, &anOID[ startPos ],
	    MIN( *extract_len, anOID_len ) * sizeof( oid ) );*/
	    for( i = startPos; i < endPos; i++ ) {
		ret[ i ] = anOID[ i ];
	    }
	    return ret;
	}
    }
    return NULL;
  
}

static void
nestTable( MDTable *table, MDTable *subTable,
	   SmiNode *smiNode, SmiNode *subNode )
{
    MDRow *iterRow;
    oid ncol = ++table->maxColNum;
    SmiElement *iterElement;
    SmiNode *idxNode = NULL;

    if( smiNode == subNode ) {
	return;
    }

    for( iterElement = smiGetFirstElement( smiNode );
	 iterElement;
	 iterElement = smiGetNextElement( iterElement ) ) {
	SmiElement *iterSubElement = smiGetFirstElement( subNode );
	SmiNode *iterSubNode = NULL;
	idxNode = smiGetElementNode( iterElement );

	for( iterSubElement = smiGetFirstElement( subNode );
	     iterSubElement;
	     iterSubElement = smiGetNextElement( iterSubElement ) ) {

	    iterSubNode = smiGetElementNode( iterSubElement );
	    if( idxNode == iterSubNode ){
		break;
	    }
	}
	if( idxNode == iterSubNode ) {
	    break;
	}
    }
    
	
    for( iterRow = subTable->rows; iterRow; iterRow = iterRow->next ) {
	MDCol *iterCol;
	size_t oidlen;
	oid *arowIndex = extractIndex( subNode, idxNode, &oidlen,
				       iterRow->index, iterRow->index_len );

	/* find the row in the parent table */
	MDRow *arow = md_findRow( table, arowIndex, oidlen );
	xfree( arowIndex );
	if ( arow ) {
	    iterCol = md_findColumn( arow, ncol );
	    if( ! iterCol ) {
		/* create a pseudo column to which the subtable
		   will be connected */
		iterCol = md_addColumn( arow, ncol, subNode->name );	    
	    }
	    if( ! iterCol->subRow ) {
		/* the first row */
		iterCol->subRow = iterRow;
	    }
	    if( ! iterCol->smiNode ) {
		iterCol->smiNode = subNode;
	    }
	}
    }
}

static int
isSubTable( SmiNode *smiNode, SmiNode *subNode )
{
    SmiElement *iterElement;
    unsigned int numIdx = numIndex( smiNode ), numSubIdx = numIndex( subNode );

    if( numSubIdx <= numIdx ) {
	return 0;
    }

    for( iterElement = smiGetFirstElement( smiNode );
	 iterElement;
	 iterElement = smiGetNextElement( iterElement ) ) {
	SmiElement *iterSubElement = smiGetFirstElement( subNode );
	SmiNode *iterSubNode;
	SmiNode *idxNode = smiGetElementNode( iterElement );
	
	for( iterSubElement = smiGetFirstElement( subNode );
	     iterSubElement;
	     iterSubElement = smiGetNextElement( iterSubElement ) ) {
	    
	    iterSubNode = smiGetElementNode( iterSubElement );
	    if( idxNode == iterSubNode ){
		return 1;
	    }
	}
    }
    return 0;
}

static void
nestAllSubTables( SmiModule *smiModule, SmiNode *smiNode, MDTable *table )
{
    SmiNode *iterNode;
    
    for( iterNode = smiGetFirstNode( smiModule, SMI_NODEKIND_ROW );
	 iterNode;
	 iterNode = smiGetNextNode( iterNode, SMI_NODEKIND_ROW ) ) {
	if( isSubTable( smiNode, iterNode ) ) {
	    MDTable *subTable = md_createTable( iterNode->name,
						oidcpy( iterNode ),
						iterNode->oidlen );
	    fillTable( subTable, iterNode, 0 );
	    nestTable( table, subTable, smiNode, iterNode );

	    /* subtables of the subtable */
	    nestAllSubTables( smiModule, iterNode, subTable );
	}
    }
}

static void
fprintElement( FILE *f, SmiNode *smiNode, oid *suffix, int suflen,
	       int subtable )
{
    struct snmp_pdu *response;
    oid *anOID;
    int oidlen;
    
    switch( smiNode->nodekind ) {
	SmiType *smiType;

    case SMI_NODEKIND_NODE :
	{
	    SmiNode *iterNode;
	    int empty = 1;
	   

	   
	    for( iterNode = smiGetFirstChildNode( smiNode );
		 iterNode;
		 iterNode = smiGetNextChildNode( iterNode ) ) {
		if( iterNode->nodekind == SMI_NODEKIND_SCALAR ) {
		    if( empty ) {
			fprintSegment( f, 1, "<%s>\n", smiNode->name );
			empty = 0;
		    }
		    fprintElement( f, iterNode, suffix, suflen, 0 );
		}
	    }

	    if( empty )
		fprintSegment( f, 0, "<%s/>\n", smiNode->name );
	    else
		fprintSegment( f, -1, "</%s>\n", smiNode->name );
	}
	break;

    case SMI_NODEKIND_TABLE :
    {
	SmiNode *iterNode;
	
	/* ignore tables and just include their entries */
	for( iterNode = smiGetFirstChildNode( smiNode );
	     iterNode;
	     iterNode = smiGetNextChildNode( iterNode ) ) {
	    fprintElement( f, iterNode, suffix, suflen, 0 );
	}
	break;
    }

    case SMI_NODEKIND_ROW :
    {
	MDTable *table;
	SmiNode *iterNode = smiNode;
	SmiModule *smiModule = smiGetNodeModule( smiNode );

	if( ( nestAugmentedTables && smiNode->indexkind != SMI_INDEX_INDEX ) )
	{
	    return;
	}

	if( nestSubtables ) {
	    SmiNode *iterTabNode;

	    for(iterTabNode = smiGetFirstNode( smiModule, SMI_NODEKIND_ROW );
		iterTabNode;
		iterTabNode = smiGetNextNode(iterTabNode, SMI_NODEKIND_ROW)) {
		if( iterTabNode != smiNode &&
		    isSubTable( iterTabNode, smiNode ) ){
		    /* we are a subnode, let's exit */
		    return;
		}
	    }
	}
	
	table = md_createTable( smiNode->name, oidcpy( smiNode ),
				smiNode->oidlen );
	fillTable( table, smiNode, 0 );

	if( nestAugmentedTables ) {
	    for( iterNode = smiGetFirstNode( smiModule, SMI_NODEKIND_ROW );
		 iterNode;
		 iterNode = smiGetNextNode( iterNode, SMI_NODEKIND_ROW ) ) {
		SmiNode *augmNode = smiGetRelatedNode( iterNode );
		if( augmNode == smiNode ) {
		    fillTable( table, iterNode, table->maxColNum );
		}
	    }
	}

	if( nestSubtables ) {
	    nestAllSubTables( smiModule, smiNode, table );
	}

	fprintTable( f, table, smiNode );
	md_dropTable( table );
	
	break;
    }
    
    case SMI_NODEKIND_SCALAR :
    {

	if( smiNode->access < SMI_ACCESS_READ_ONLY ) {
	    return;
	}
	
	smiType = smiGetNodeType( smiNode );

	/* get values from agent*/
	anOID = oidcpy( smiNode );
	anOID = oidadd( anOID, smiNode->oidlen, suffix, suflen, &oidlen );
	
	if( snmpGet( anOID, oidlen, SNMP_MSG_GET, &response ) ) {
	    char *strval = getStringVal( response->variables, smiNode );
	    if( strval ) {
		fprintSegment( f, 0, "<%s>%s</%s>\n",
			       smiNode->name, strval, smiNode->name );
	    }
	}
	else if( !silentFlag ) {
	    fprint( f, "<!-- element %s not found on agent -->\n",
		    smiNode->name );
	}
	xfree( anOID );
	break;
    }

    case SMI_NODEKIND_NOTIFICATION:
	/* this should not occur here */
	break;

    default:
	fprint( f,
		"<!-- Warning! Unhandled Element! No details available!\n" );
	fprint( f, "      Nodekind: %#4x -->\n\n" );
	
    }
}


static void
fprintRows( FILE *f, SmiModule *smiModule )
{
    SmiNode *iterNode;

    for( iterNode = smiGetFirstNode( smiModule, SMI_NODEKIND_ROW );
	 iterNode;
	 iterNode = smiGetNextNode( iterNode,  SMI_NODEKIND_ROW ) ) {
	if( hasChildren( iterNode,
			 SMI_NODEKIND_COLUMN | SMI_NODEKIND_TABLE ) ){
	    
	    fprintElement( f, iterNode, NULL, 0, 0 );
	}
    }
}


static void
fprintNodes( FILE *f, SmiModule *smiModule )
{
    SmiNode *iterNode;
    oid *suffix = xmalloc( sizeof( oid ) );

    suffix[0] = 0;
    
    for( iterNode = smiGetFirstNode( smiModule, SMI_NODEKIND_NODE );
	 iterNode;
	 iterNode = smiGetNextNode( iterNode, SMI_NODEKIND_NODE ) ) {
	if( hasChildren( iterNode, SMI_NODEKIND_SCALAR ) ) {
	    fprintElement( f, iterNode, suffix, 1, 0 );
	}
    }


    
    xfree( suffix );
}


static void
fprintModule(FILE *f, SmiModule *smiModule)
{
    /* get actual time stamp */
    time_t tmstamp = time( NULL );
    struct tm *tmst = gmtime( &tmstamp );
    char strtm[25];
    struct hostent *hostinfo = gethostbyname( snmpHost );
    unsigned int i;
    char *ipaddr = NULL;
    
    strftime( strtm, 25, "%G-%m-%dT%H:%M:%SZ", tmst );
    
    if( hostinfo ) {
	for( i = 0; i < hostinfo->h_length; i++ ) {
	    unsigned short int ap = (unsigned char)hostinfo->h_addr_list[0][i] ; 
	    asprintf( &ipaddr, "%s%i", ipaddr ? ipaddr : "", ap );
	    if( i < hostinfo->h_length - 1 ) {
		asprintf( &ipaddr, "%s.", ipaddr ); 
	    }
	}
	asprintf( &ipaddr, "%s\0", ipaddr );
    }
    
    /* print root node */
    fprintf( f, "<snmp-data xmlns=\"%s%s\"\n",
	     schemaLocation, smiModule->name );
/*    fprintSegment( f, strlen( smiModule->name ) + 2,
		   "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n",
		   0 );
*/
    fprintf(f, "   xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n");
    
    fprintf( f, "xsi:schemaLocation=\"%s%s %s%s.xsd\">\n",
	     schemaLocation, smiModule->name,
	     schemaLocation, smiModule->name );
    if( hostinfo ) {
	fprintSegment( f, 1, "<context ipaddr=\"%s\" ", ipaddr );
	if( hostinfo->h_name ) {
	    fprintf( f, "hostname=\"%s\" ", hostinfo->h_name );
	}
	fprintf( f, "port=\"%d\" community=\"%s\" time=\"%s\">\n",
		 snmpPort, snmpCommunity, strtm );
    }
    /* xxx error handling */
       
    /* print other nodes (scalars and scalar groups) */
    fprintNodes( f, smiModule );
    
    /* print tables */
    fprintRows( f, smiModule );
    fprintSegment( f, -1, "</context>\n" );
    
    fprintf( f, "</snmp-data>\n" );
}

static void
dumpXml( SmiModule *module, int flags )
{
    FILE *f = stdout;

    silentFlag = flags & MIBDUMP_FLAG_SILENT;

    /* make sure url ends with '/' */
    if( schemaLocation[ strlen( schemaLocation ) - 1 ] != '/' ) {
	strcat( schemaLocation, "/" );
    }
	
    fprint(f, "<?xml version=\"1.0\"?>\n"); 
    fprint(f, "<!-- This module has been generated by mibdump "
	   MIBDUMP_VERSION_STRING ". Do not edit. -->\n");
    fprint(f, "\n");

    fprintModule( f, module );
}


void
initXml()
{

    static MibdumpDriverOption opt[] = {
	{ "schema-url", OPT_STRING, &schemaLocation, 0,
	  "The URL where the imported MIB modules reside" },
	{ "units", OPT_FLAG, &dumpUnits, 0,
	  "Dump unit definitions as attributes" },
	{ "hints", OPT_FLAG, &dumpDisplayHint, 0,
	  "Dump display hints as attributes" },
	{ "nest-augments", OPT_FLAG, &nestAugmentedTables, 0,
	  "Nest rows of augmented tables in the base tables" },
	{ "nest-subtables", OPT_FLAG, &nestSubtables, 0,
	  "Nest subtables in the base tables" },
	{ 0, OPT_END, 0, 0 }
    };
    
    static MibdumpDriver driver = {
	"xml",
	dumpXml,
	0,
	MIBDUMP_DRIVER_CANT_UNITE,
	"XML format",
	opt,
	NULL
    };
    
    mibdumpRegisterDriver(&driver);
}
